﻿using System;
using System.Collections.Generic;
using System.Windows.Input;
using WpfGeoTagger.Commands;
using WpfGeoTagger.Models;
using System.Xml.Linq;
using System.Xml;
using System.Diagnostics;

namespace WpfGeoTagger.ViewModels
{
	/// <summary>
	/// Map の Model と View を仲介します。
	/// </summary>
	sealed class MapViewModel : ViewModelBase
	{
		/// <summary>
		/// インスタンスを初期化します。
		/// </summary>
		/// <param name="uri">ブラウザに表示するページの URI。</param>
		/// <param name="invokScript">ブラウザのスクリプト実行を行う為のデリゲート。</param>
		public MapViewModel( string uri )
		{
			this.Uri                    = uri;
			this.MapLocation               = new GeoTag();
			this.MapHost                = new MapHost();
			this.MapHost.MapChanged    += new EventHandler< MapChangedEventArgs >( OnMapChanged );
			this.MapHost.MarkerChanged += new EventHandler< MarkerChangedEventArgs >( OnMarkerChanged );

			// 編集結果の保存
			this.EditSaveCommand = new DelegateCommand(
				() =>
				{
					this._selectedImage.SaveGeoTag();
				},
				() =>
				{
					return ( this._selectedImage != null && this._selectedImage.IsEdited );
				}
				);

			// マーカー追加
			this.AddMarkerCommand = new DelegateCommand(
				() =>
				{
					this.AddMarker( this._selectedImage );
				},
				() =>
				{
					return ( this._selectedImage != null && this._selectedImage.GeoTag == null );
				}
				);

			// マーカー削除
			this.RemoveMarkerCommand = new DelegateCommand(
				() =>
				{
					this._selectedImage.GeoTag = null;
					this.InvokeScript( "removeMarker" );
				},
				() =>
				{
					return ( this._selectedImage != null && this._selectedImage.GeoTag != null );
				}
				);

			// マーカー位置の表示
			this.ShowMarkerCommand = new DelegateCommand(
				() =>
				{
					this.InvokeScript( "showMarker" );
				},
				() =>
				{
					return ( this._selectedImage != null && this._selectedImage.GeoTag != null );
				}
				);
			
			// マップ移動
			this.MoveMapCommand = new DelegateCommand( () => { this.InvokeScript( "moveMap", this.Address ); } );
		}

		/// <summary>
		/// 指定された画像の GeoTag をマーカーとして追加します。
		/// </summary>
		/// <param name="image"></param>
		public void AddMarker( ImageViewModel image )
		{
			// 画像に GeoTag がない場合は、マップの中央位置を設定する
			if( image.GeoTag == null )
			{
				image.GeoTag = new GeoTag() { Latitude = this.MapLocation.Latitude, Longitude = this.MapLocation.Longitude };
			}

			this._images.Add( image.Id, image );

			this.InvokeScript( "addMarker", image.Id, image.Name, image.GeoTag.Latitude, image.GeoTag.Longitude );
		}

		/// <summary>
		/// 指定された識別子が現在選択されている画像を示している事を調べます。
		/// </summary>
		/// <param name="id">識別子。</param>
		/// <returns>選択されている画像の場合は true。それ以外は false。</returns>
		private bool IsSelectedId( int id )
		{
			return ( this._selectedImage != null && this._selectedImage.Id == id );
		}

		/// <summary>
		/// マップが変更された時に発生するイベントです。
		/// </summary>
		/// <param name="sender">イベント送信元。</param>
		/// <param name="e">イベント データ。</param>
		private void OnMapChanged( object sender, MapChangedEventArgs e )
		{
			this.MapLocation.Latitude  = e.Latitude;
			this.MapLocation.Longitude = e.Longitude;
			this.NotifyPropertyChanged( "MapLocation" );
		}

		/// <summary>
		/// マーカーが変更された時に発生するイベントです。
		/// </summary>
		/// <param name="sender">イベント送信元。</param>
		/// <param name="e">イベント データ。</param>
		private void OnMarkerChanged( object sender, MarkerChangedEventArgs e )
		{
			switch( e.Action )
			{
			// 追加
			case MarkerAction.Added:
				{
					var image = this._images[ e.Id ];
					image.GeoTag.UpdateAltitude();

					if( this._selectedImage == null )
					{
						this.SelectedImage = image;
					}
					else if( this.IsSelectedId( e.Id ) )
					{
						this.NotifyPropertyChanged( "MarkerLocation" );
					}
				}
				break;

			// 移動
			case MarkerAction.Moved:
				{
					var image = this._images[ e.Id ];
					image.GeoTag.Latitude  = e.Latitude;
					image.GeoTag.Longitude = e.Longitude;

					if( this.IsSelectedId( e.Id ) )
					{
						this.NotifyPropertyChanged( "MarkerLocation" );
					}
				}
				break;

			// 移動完了
			case MarkerAction.MoveFinished:
				{
					var image = this._images[ e.Id ];
					image.GeoTag.UpdateAltitude();
					
					if( this.IsSelectedId( e.Id ) )
					{
						this.NotifyPropertyChanged( "MarkerLocation" );
					}
				}
				break;

			// 移動開始
			case MarkerAction.MoveStarted:
				{
					// 移動が完了するまで標高を無効化する
					var image = this._images[ e.Id ];
					image.GeoTag.Altitude = 0.0;

					if( this.IsSelectedId( e.Id ) )
					{
						this.NotifyPropertyChanged( "MarkerLocation" );
					}
				}
				break;


			// 削除
			case MarkerAction.Removed:
				this._images.Remove( e.Id );

				if( this.IsSelectedId( e.Id ) )
				{
					this.SelectedImage = null;
				}

				break;

			// 選択
			case MarkerAction.Selected:
				this.SelectedImage = this._images[ e.Id ];
				break;
			}
		}

		#region プロパティ

		/// <summary>
		/// ブラウザのスクリプト実行を行う為のデリゲート。
		/// </summary>
		/// <param name="scriptName">実行するスクリプト関数の名前。</param>
		/// <param name="args">スクリプト関数に渡すパラメータ。</param>
		public delegate void InvokeScriptHandler( string scriptName, params object[] args );

		/// <summary>
		/// ブラウザのスクリプト実行を行う為に呼び出されるメソッドを取得または設定します。
		/// </summary>
		public event InvokeScriptHandler InvokeScript;

		/// <summary>
		/// マップに指定する住所を取得または設定します。
		/// </summary>
		public string Address { get; set; }

		/// <summary>
		/// マップ中央の緯度・経度を示す文字列を取得または設定します。
		/// </summary>
		public GeoTag MapLocation { get; private set; }

		/// <summary>
		/// マーカーの緯度・経度を示す文字列を取得または設定します。
		/// </summary>
		public GeoTag MarkerLocation { get { return ( this._selectedImage == null ? null : this._selectedImage.GeoTag ); } }

		/// <summary>
		///  マップの操作を行う JavaScript に関連付けられるオブジェクトを取得します。
		/// </summary>
		public MapHost MapHost { get; private set; }

		/// <summary>
		/// 選択されている画像を取得または設定します。
		/// </summary>
		public ImageViewModel SelectedImage
		{
			get
			{
				return this._selectedImage;
			}
			set
			{
				this._selectedImage = value;

				if( this._selectedImage != null )
				{
					this.InvokeScript( "selectMarker", this._selectedImage.Id );
				}

				this.NotifyPropertyChanged( "SelectedImage", "MarkerLocation" );
			}
		}

		/// <summary>
		/// ブラウザに表示するページの URI を取得します。
		/// </summary>
		public string Uri { get; private set; }

		#endregion

		#region コマンド

		/// <summary>
		/// マーカーの追加コマンドを取得します。
		/// </summary>
		public ICommand AddMarkerCommand { get; private set; }

		/// <summary>
		/// 編集内容の保存コマンドを取得します。
		/// </summary>
		public ICommand EditSaveCommand { get; private set; }

		/// <summary>
		/// マーカーの削除コマンドを取得します。
		/// </summary>
		public ICommand RemoveMarkerCommand { get; private set; }

		/// <summary>
		/// マーカーの表示を行うコマンドを取得します。
		/// </summary>
		public ICommand ShowMarkerCommand { get; private set; }

		/// <summary>
		/// 現在設定されている Address の位置にマップを移動させるコマンドを取得します。
		/// </summary>
		public ICommand MoveMapCommand { get; private set; }

		/// <summary>
		/// マーカーの位置へマップを移動させるコマンドを取得します。
		/// </summary>
		public ICommand MoveMapToMarkerCommand { get; private set; }

		#endregion

		#region フィールド

		/// <summary>
		/// 識別子をキーとする画像情報のディクショナリ。
		/// </summary>
		private Dictionary< int, ImageViewModel > _images = new Dictionary< int, ImageViewModel >();

		/// <summary>
		/// 選択されている画像。
		/// </summary>
		private ImageViewModel _selectedImage;

		#endregion
	}
}
